"""Top-level package for aiolifxc."""

from .aiolifx import Devices, Lights, Device, Light, LifxDiscovery, DeviceOffline  # NOQA
from .colors import Color  # NOQA

__author__ = """Brian May"""
__email__ = 'brian@linuxpenguins.xyz'
__version__ = '0.5.3'
